/*
    $Header: gnuuio.c,v 1.45 90/02/28 04:05:55 gnu Exp $
    Copyright 1989 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 * Derived from:
 * i[$]uuslave.c	1.7 08/12/85 14:04:20
 * which came from the ACGNJ BBS system at +1 201 753 9758.  Original
 * author unknown.  Many people, too numerous to list here, have contributed
 * changes and ports to this program.
 */

#ifndef lint
char version[] = "hoptoad $Revision: 1.45 $";
#endif

/*

This program implements the uucp (Unix-to-Unix CoPy) protocol, used to
transfer mail, files, and Usenet news among Unix machines.  UUCP comes
free with Unix (unless you get a sleazeoid version like Xenix, where
they charge you extra for it), but you don't get sources.  You can buy
a commercial program for MSDOS, called UULINK, which also implements
this protocol.  UULINK costs $300 and you still don't get sources.
This program not only runs on Unix and MSDOS, but on many other
machines, and comes with the Free Software Foundation copyright, which
guarantees you access to sources.

The protocol requires a full 8-bit data path with no characters inserted
or deleted (e.g. ^S and ^Q are used as DATA characters).  Simple serial
ports and modems do this; most complicated networks do not, at least without
setting up odd modes and such.  Telenet's PC Pursuit works fine though.

The basic flow of the protocol is that the calling machine will send down
a line of text saying what it wants to do (send a file, receive a file,
or hang up).  (The lines of text are encapsulated into packets; see below.)
The called machine responds with a "yes" or "no" answer, and if the answer
was yes, it sends or receives the file.  Files are terminated with a
packet containing 0 bytes of data.  Then the system that received the file
sends a "copy succeeded" or "copy failed" line to the other end, and 
they go back to "what do we do now".  A request to hang up should be
answered "no" if the called machine has some mail or files it wants to
send to the calling machine; the two machines reverse roles and the calling
machine goes into "what do we do now".  If a hangup request is answered "yes",
the call is terminated.

The data flow described above is actually sent by a lower level "protocol
module".  The default "g" protocol module sends the data in packets containing
checksums and acknowledgements.  Each packet can either hold a short
control message, e.g. an ack, or a data block.  The data blocks are
numbered with a 3-bit counter, and sent with a checksum.  If the sender
has not received an acknowledgement for a data block within a certain
time, it retransmits the block.  The size of a data block is negotiated
at the start of a call.  To send a block with fewer bytes, a "short
data" block is sent, which is just as big as a "long data" block, but
contains a 1- or 2-byte count of "how many bytes in this block are just
padding".  This is a cute trick since it always works (e.g. if you want
to send 1023 out of 1024 bytes, you only need one byte for the count;
while if you want to send 1 byte out of 1024 then you have enough space
for the count to be 2 bytes).

The short control messages are used to start the call and negotiate the
packet size and the "window size", to acknowledge or reject packets,
and to terminate the packet protocol at the end of a call.  The window
size is how many packets one side can send before it will stop and wait
for an acknowledgement from the other side.  A window size of 1 makes
for a half-duplex protocol (which is what gnuucp currently
implements), but also makes it easy to implement on micros that don't
handle serial lines with interrupts.  In window 1, you just keep
sending the same packet until the other side acknowledges it.  Unix
always uses a window size of 3, which is the max that can be dealt with
given the 3-bit packet numbers (for reasons that would take more space
than I want to spend here).  This gives much better throughput, but
requires full duplex serial port handling and more complicated
acknowledgement strategies.

In receiving data, the "g" protocol scans for a DLE (hex 10) which
indicates the start of a packet, and the next 5 bytes are read and
checked.  If they pass, this is a good packet and it is acted upon.
(If it's a data packet, we have to read in and check the data part
too.)  If the checks fail, all the bytes read so far should be rescanned
for another DLE, since it's possible that some characters were dropped
and the first DLE we saw was actually part of a data packet.

Other protocol modules have been implemented; e.g. the "f" protocol which
eliminates the packets and acknowledgements, for reduced overhead when
running over an error-free, flow controlled connection on an X.25 network.
There has been talk of a "z" protocol which implements ZMODEM.  So far,
none of these modules are supplied with gnuucp, though "f" is public domain
and "z" is intended to be, because they are not ported and tested yet.

At the start of a call, the caller sends the list of protocols it supports,
and the callee picks the best one it knows and says "use this one please".

At the low level, full 8-bit bytes are sent out and received on an async
serial port.
*/

#include "includes.h"		/* System include files, system dependent */
#include "uucp.h"		/* Uucp definitions and parameters */
#include "sysdep.h"		/* System dependent parts of gnuucp */

#define	MAX_FLAGS	40

extern int errno;

extern int fdtty;			/* Opaque handle to pass to protocol */

char	ttynam[NAMESIZE],		/* Name of tty we use as serial port */
	srcnam[NAMESIZE],		/* Source file name */
	dstnam[NAMESIZE],		/* Dest file name */
	flags[MAX_FLAGS],		/* Flags from file xfer cmd */
	temp[NAMESIZE];			/* Temp file name */

int	ignore_time_restrictions = 0,	/* Call out even if L.sys sez no */
	mode;				/* File mode from file xfer cmd */

/* Flag bits for flags -- FIXME -- currently unused. */
#define	F_COPY	1			/* -c or -C:  copy? */
#define	F_DIRS	2			/* -f or -d */
#define	F_NOTIFY	4		/* -n:  notify recipient */
#define	F_MAIL		8		/* -m:  notify originator */

int   f_wait    = 0;	/* FIXME, unreferenced now */
int   loop    = 1;	/* Loop accepting logins if tty name specified */

/*
 * Status of the comm line -- passed as results from subroutines
 * that handle the flow of commands over the line.
 */
typedef enum {
	WSUCCESS =	0,		/* Protocol is working */
	WFAIL =		1,		/* Protocol failed */
	HANGUP =	2,		/* Switch master/slave roles */
	HANGNOW =	3,		/* Hang up now */
	COPYFAIL =	4,		/* File copy failed */
} line_status;

s_or_f yesno();
void logit();
void do_session();
void xlat_str();
s_or_f call_sysline();
s_or_f top_level();
extern s_or_f openout();
void call_system();
line_status do_one_slave(), do_one_master();
line_status host_send_file(), host_receive_file();
line_status local_send_file(), local_receive_file();
line_status send_file(), receive_file();
extern s_or_f read_params();

#define	PROT_RW_OWNER	0600	/* Protection mask allowing owner read&write */

#define	MAX_STRING	200	/* Max length string to send/expect */

/* Strings sent by the slave side to the master */
static char msgo0[] = "login: ";
static char msgo1[] = "Password:";
static char msgo3[] = "\20ROK\0";
static char msgo4[] = "\20OOOOOOO\0";

/* Strings sent by the master to the slave */
static char msgi0[] = "uucp\n";
static char msgi1[] = "s8000\n";
static char msg_noproto[] = "\20UN\0";	/* No protocol in common */
static char msgi4[] = "OOOOOO";

/*
 * Protocol switch data structure
 */
struct proto {
	char	p_id;
	s_or_f	(*p_turnon)();
	s_or_f	(*p_rdmsg)();
	s_or_f	(*p_wrmsg)();
	s_or_f	(*p_rddata)();
	s_or_f	(*p_wrdata)();
	s_or_f	(*p_turnoff)();
};

extern s_or_f gturnon(), grdmsg(), gwrmsg(), grddata(), gwrdata(), gturnoff();

/*
 * Table of supported protocols.
 * ORDER IS SIGNIFICANT!! from most desirable to least.
 * When a protocol becomes supported, add its function declarations
 * above, and add the function names to the table below.
 */
struct proto ptbl[] = {
/* "t" protocol: BSD TCP error free protocol */
/* "e" protocol: HDB error-free protocol */
/* "d" protocol: HDB datakit protocol */
/* "f" protocol to run on BSD X.25 PADs */
/* "x" protocol: HDB S5R2 X.25 protocol */
/* "u" protocol: GNU UDP packet protocol */
/* "z" protocol: zmodem transfer protocol (pyramid!csg and uunet!rick) */
/* "G" protocol: SVR4 "g" that handles w=7 and full segs */
/* "g" protocol for dialup async modem lines */
	{'g', gturnon, grdmsg, gwrmsg, grddata, gwrdata, gturnoff},
/* "h" protocol: GNU packet protocol for 7bit datapaths */
/* "k" protocol: "kermit"-like protocol (several versions?)*/
	{'\0', 0, 0, 0, 0, 0, 0},
};

struct proto *curproto = &ptbl[0];

#define	turnon	(*curproto->p_turnon)
#define	rdmsg	(*curproto->p_rdmsg)
#define	wrmsg	(*curproto->p_wrmsg)
#define	rddata	(*curproto->p_rddata)
#define	wrdata	(*curproto->p_wrdata)
#define	turnoff	(*curproto->p_turnoff)

/*
 * Read from the serial port a null-terminated string.
 * The string FIXME should not overrun MAX_STRING.
 * String starts with first printing character.
 */
s_or_f
getstring(where)
	register char *where;
{
	register int data, count = 0;

	DEBUG(2, "Getstring:  ", 0);
	/* Read data until null character */
	while ( ((data = xgetc()) != EOF)) {
		data &= 0x7F;
		if (DEBUG_LEVEL(2)) {
			DEBUG_DATA(data);
		}

		where[count++] = data;
		if (data == 0x00) {
			if (DEBUG_LEVEL(2)) putchar('\n');
			return SUCCESS;
		}
		if (count == 1 && !isprint(where[0]))
			count = 0;
	}
	if (DEBUG_LEVEL(2)) putchar('\n');
	return FAIL;
}


/*
 * Medium level input routine.
 *
 * Look for an input string for the send-expect sequence.
 * Return SUCCESS for matching string, FAIL for timeout before we found it.
 * FIXME:  we only time out if the other end stops sending.  If it
 *	   keeps sending, we keep listening forever.
 */
s_or_f
instr(s,n)
register char *s;
register int n;
{
	register int data;
	register char	*pattern;
	register int i;

	if (DEBUG_LEVEL(2)) {
		printf("Expecting ");
		for (i = 0; i < n; i++)
			DEBUG_DATA(s[i]);
		printf("\nR ");
	}

	pattern = s;

	while ((data = xgetc()) != EOF) {
		data &= 0x7F;			/* Ignore parity */

		if (DEBUG_LEVEL(2)) {
			DEBUG_DATA(data);
		}

	tryfirst:

		if (data == *pattern) {
			/* Input matches next byte of pattern */
			pattern++;
			if (pattern >= &s[n]) {
				if (DEBUG_LEVEL(2)) putchar('\n');
				return SUCCESS;		/* Done */
			}
		} else {
			/* Input doesn't match pattern */
			if (pattern == s) continue;	/* First char */
			/* FIXME!  This doesn't work if pattern can match
			 * in more than one way initially, e.g.:
			 * "NECT" works when "CONNECT" is output, but
			 * "NNECT" fails when "CONNNECT" is output... */
			pattern = s;	/* Does cur char match first? */
			goto tryfirst;	/* FIXME */
		}
	}

	if (DEBUG_LEVEL(2))	putchar('\n');
	DEBUG(1, "Didn't see expected string '%s'\n", s);
	return FAIL;
}

/*
 * Debugging hack for stuff written to the modem.
 */
int
twrite(s, n)
	char *s;
	int	n;
{
	register int i;

	if (DEBUG_LEVEL(2)) {
		printf("Wrote:  ");
		for (i = 0; i < n; i++)
			DEBUG_DATA(s[i]);
		printf("\n");
	}

	return xwrite(s, n);
}


/* Debug routines for printing serial port data. 
 *
 * debug_data_easy is used until we switch into protocol mode.
 */
void
debug_data_easy(data)
	char data;
{
	if (isprint(data) || isspace(data))	
		putchar(data);
	else
		printf("\\x%02x", data & 0xFF);
}

void
debug_data_ugly(data)
	char data;
{
	printf("%02x%c ", data & 0xFF, (isascii(data) && isprint(data))?
				       data: ' ');
}

/*
 * MAIN ROUTINE.
 *
 * This is called at program startup.  It parses the arguments to the
 * program (if any) and sets up to receive a call on the modem.
 *
 * If there are no arguments, we assume the caller is already on standard
 * input, waiting to do uucp protocols (past the login prompt), and we
 * just handle one caller.
 *
 * If there is an argument, it is the name of the tty device where we
 * should listen for multiple callers and handle login and password.
 */
void
main(argc,argv)
int argc;
char *argv[];
{
	int	ontheline = 1;
	int	i;
	char	*poll_sys = (char *)NULL;  /* System name to poll, or none */

#ifdef GNUUIO_MAIN_HOOK
	GNUUIO_MAIN_HOOK;
#endif

#ifdef MSC
	/* Turbo C doesn't seem to be able to handle signal()...I guess
	 * you are just supposed to reboot the system when you want it back.
	 */
	if (signal(SIGINT,handler) == (int(*)())-1) {
		fprintf(stderr,"Couldn't set SIGINT\n");
		abort();
	}
#endif

	debug = -1;		/* Let read_params set it if -x doesn't */
	Debug_data = debug_data_easy;	/* In case it gets called */

	/* FIXME, use getopt */
	/* scan command line arguments, kinda kludgy but it works */
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			break;
		switch (argv[i][1]) {

		case 'w':
			f_wait++;
			printf("gnuuio: will wait for call after outbound\n");
			break;

		case 'x':
			debug = atoi(&argv[i][2]);
			break;

		case 'S':
			ignore_time_restrictions++;
		case 's':
			poll_sys = &argv[i][2];
			ontheline = 0;
			break;

		case 'e':
			loop++;
			printf("gnuuio: endless loop mode\n");
			break;

		case 'C':
			Control = &argv[i][2];
			break;

		default:
			printf(
			  "gnuuio: I don't recognize -%c as a valid switch.\n",
				argv[i][1]);
			break;
		}
	}

	/* If argument provided, use it as name of comm port */
	if (i < argc) {
		ontheline = 0;
		strcpy(ttynam, argv[i]);
	}

	read_params(Control);		/* Read control file */

	if (chdir(Spool)) {
		fprintf(stderr, "Can't chdir to Spool dir: ");
		perror(Spool);
		exit(2);
	}

	/*
	 * If running via getty/login, our debug stdout had better
	 * go to a file, not to the usual stdout!
	 */
	if (DEBUG_LEVEL(0) && ontheline) {
		freopen("gnuucp.log", "a", stdout);
	}

	setbuf(stdout, (char *)NULL);	/* Unbuffered debug output */

	/* Timestamp the long debug log */
	if (DEBUG_LEVEL(0)) {
		printf("\n\n\n\ngnuucp log on tty '%s' starting %s\n",
			ttynam, time_and_pid());
	}

	if (poll_sys) {
		if (*poll_sys == '\0') poll_sys = (char *)NULL;
		call_system(poll_sys);
		if (!f_wait) goto end;
	}

	do {
		/*
		 *  Set up serial channel, wait for incoming call. 
		 */

		DEBUG(0, "\nRestarting\n", 0);
		openline(ttynam, (long)1200);	/* FIXME, let user set baud */

		do_session(ontheline);

		hangup();
		DEBUG(0, "\nEnd of call\n", 0);

	} while (loop && !ontheline);

end:
	exit_gnuucp(0, 1);		/* Clean exit, returning 0 */
}

/*
 * Handle a single set of uucp expect/send strings.
 *
 * Return SUCCESS if the expect string is found, or FAIL if not.
 */
s_or_f
sendexpect(expect, send)
	register char *expect;
	char *send;
{
	char *dasher, *dancer;

	while ((dasher = strchr(expect, '-')) != NULL
	    && (dancer = strchr(dasher+1, '-')) != NULL) {
		/* Two dashes found -- we have an expect-send-expect string */
		*dasher = '\0'; *dancer = '\0';
		if (instr(expect, dasher-expect) == SUCCESS)
			goto sendit;
		xlat_str(dasher+1);		/* Send the alternative */
		expect = dancer + 1;		/* Try next pair, if any */
	}
		
	if (expect[0] != '"' || expect[1] != '"' || expect[2] != '\0')
		if (instr(expect, strlen(expect)) != SUCCESS)
			return FAIL;

sendit:
	if (send) {
		xlat_str(send);
	}
	return SUCCESS;
}


/*
 * translate embedded escape characters in a "send" string, and send 'em.
 */
void
xlat_str(msg)
	register char	*msg;
{
	register int	i  = 0,
		j  = 0;
	int	cr = 1;
	char	out[MAX_STRING+SLOP];

	sleep(1);		/* Old uucp did this, guess we'd better */
	/* Handle EOT and BREAK here...FIXME */
	while (msg[i]) {
		if (msg[i] == '\\') {
			switch (msg[++i]) {
			case 'r':		/* carriage return */
				out[j++] = 0x0d;
				break;
			case 'n':		/* line feed */
				out[j++] = 0x0a;
				break;
			case '\\':		/* back slash */
				out[j++] = '\\';
				break;
			case 't':		/* tab */
				out[j++] = '\t';
				break;
			case 'B':		/* HDB: backspace */
				out[j++] = 0x08;
				break;
			case 'N':		/* HDB: null */
				out[j++] = 0x00;
				break;
			case 'p':		/* HDB: pause (.5 sec) */
			case 'd':		/* delay */
				/* Print output, delay, resume */
				twrite(out, j);
				j = 0;
				if (DEBUG_LEVEL(2)) printf("(SLEEP)  ");
				sleep(1);
				break;
			case 's':		/* space */
				out[j++] = ' ';
				break;
			case 'c':		/* no CR at end */
				cr = 0;
				break;
			case 'K':		/* HDB: BREAK (in dialers?) */
			case 'b':		/* break */
				twrite(out, j);	/* First kick out output */
				j = 0;
				if (DEBUG_LEVEL(2)) printf("(BREAK)  ");
				sendbreak();
				break;
			default:		/* don't know so skip it */
				/* FIXME:  handle \ooo with octal */
				break;
			}
			i++;
		} else if (msg[i] == 'B' && strcmp(msg, "BREAK") == SAME) {
			if (DEBUG_LEVEL(2)) printf("(BREAK)  ");
			sendbreak();
			return;
		} else if (msg[i] == 'E' && strcmp(msg, "EOF") == SAME) {
			out[j++] = 0x04;
			goto eggzig;
		} else {
			out[j++] = msg[i++];
		}
	}

	if (cr)
		out[j++] = 0x0d;
eggzig:
	if (j) twrite(out, j);
}

/*
 * Line finder.  Given an "ACU" specifier from the L.sys file,
 * and a baud rate, determine a file name that we should try to lock
 * and open.
 *
 * On initial call, pass in (struct port *)0 as the port argument.
 * On subsequent calls, we skip past that port, which is presumed to
 * have been tried and failed.
 *
 * If we find a matching entry in the config file ports database,
 * return it; else return a null pointer.
 */
struct port *
pickport(acu, baud, p)
	char	*acu;
	long	baud;
	register struct port *p;
{

	if (p == 0)
		p = ports;
	else
		p = p->chain;

	for ( ; p; p = p->chain) {
		if (baud == p->baud &&
		    strcmp(acu, p->portname) == 0)
			return p;
	}

	return p;			/* Didn't find it -- return null ptr */
}


/*
 * Simple dialer routine.  Needs replacement with a full blown
 * script driven dialer.  Next week maybe :-).  FIXME.
 */
s_or_f
dial_nbr(port, nbr)
	struct port *port;
	char  *nbr;
{
	char  dial[30];
	char  conn[30];
	int   i;
	s_or_f n;

	/* Hardwired lines do no dialing. */
	if (strcmp("none", port->modemname) == 0)
		return SUCCESS;

	/* If it doesn't say hayes, we can't dial it. */
	if (strcmp("hayes", port->modemname) != 0)
		return FAIL;

	sprintf(dial, "ATV1DT %s\r", nbr);

	if (DEBUG_LEVEL(2))
		printf("dialing %s at %ld baud: %s\n",
			port->devname, port->baud, dial);

	twrite(dial, strlen(dial));

	/* FIXME, we've got to be more flexible here... */
	sprintf(conn, "CONNECT");

	n = instr(conn, strlen(conn));

	/* FIXME, scan out the baud rate, or whatever's before the CR, here.
	 */

	if (n == SUCCESS) sleep(2);	/* Let the line settle. */

	return (n);
}

/*
 * Call a specific system, or all systems that have work pending.
 */
void
call_system(sys)
	char	*sys;
{
	FILE	*lsys;
	char	buf[MAX_LSYS];
	char	sysnam[MAX_HOST];
	char	prev_name[MAX_HOST];
	s_or_f	called = FAIL;

	/*
	 * Unix uucico just reads the directory, and calls the systems
	 * in the order of the files in the directory.  We want more
	 * control than that, though I'm not sure that L.sys order is
	 * best either.  For example, in the first call after 11PM,
	 * I'd like to call the sites that haven't been callable before
	 * 11PM first, and finish up with the ones I've been able to call
	 * all day.  FIXME.
	 */

	if (! (lsys = fopen(Sysfile, "r"))) {
		DEBUG(0, "gnuuio: can't open %s", Sysfile);
		DEBUG(0, ", errno %d\n", errno);
		return;
	}
	sysnam[0] = '\0';		/* Initially, no previous sys */

	/* Once per system in L.sys... */
	/* FIXME, handle continuation lines (trailing "\") */
	while (fgets(buf, sizeof buf, lsys)) {

		if (buf[0] == '#')
			continue;
		
		/*
		 * Grab the system name.  If same as previous, and
		 * the previous call worked, skip it.
		 */
		strcpy(prev_name, sysnam);
		(void) sscanf(buf, "%s", sysnam);
		if (!strcmp(sysnam, prev_name)) {
			if (called == SUCCESS) continue;
		}

		/*
		 * If a system name was specified, skip til we find it
		 * If none was specified, only call if there is work.
		 */
		if (sys) {
			if (0 != strcmp(sys, sysnam))
				continue;
		} else {
			DEBUG(4,"searching for outbound work for %s\n", sysnam);

			if (!work_scan(sysnam, "C")) {
				DEBUG(3,"no work for %s\n", sysnam);
				called = SUCCESS;	/* Don't try further */
				continue;
			}

			DEBUG(3, "found work for %s\n", sysnam);
		}

		called = call_sysline(buf);

		/* If it worked and we only wanted to call one system, quit */
		if (called == SUCCESS && sys) break;
	}

	fclose(lsys);
	if (called == FAIL && sys)
		DEBUG(0, "Could not call system %s\n", sys);
	return;
}

/*
 * Call out to a system, given its L.sys line.
 */
s_or_f
call_sysline(lsysline)
	char	*lsysline;
{
	char	tempname[MAX_HOST + 30 + SLOP],
		*sysnam,
		*times,
		*acu,
		*sbaud,
		*telno,
		*send,
		*expect,
		*protocols;
	struct port *port;
	char	logbuf[MAX_HOST+30+SLOP];
	char	msgbuf[MAX_STRING+SLOP];
	long	baud;
	static char delims[] = " \t";	/* L.sys field delimiters */
	register struct proto *pp;

	who[0] = '-'; who[1] = '\0';	/* No user now (for logit) */

	/* FIXME, use the values it is ignoring here */
	sysnam= strtok(lsysline,     delims);	/* System name */
	times = strtok((char *)NULL, delims);	/* Time */
	acu   = strtok((char *)NULL, delims);	/* ACU */
	sbaud = strtok((char *)NULL, delims);	/* Baud */
	telno = strtok((char *)NULL, delims);	/* phone */

	strcpy(host_name, sysnam);

	/*
	 * Check the time parameter and return FAIL if
	 * it does not allow calls now, unless overriden with -S.
	 */
	if (!ignore_time_restrictions) {
		if (!is_time_ok(times)) {
			logit("WRONG TIME TO CALL", sysnam);
			return FAIL;
		}
	}

	baud = atol(sbaud);

	protocols = strchr(acu, ',');
	if (protocols != NULL) {	/* have protocol list */
		*protocols++ = '\0';	/* tie off acu */

		/* See if we support any of the specified protocols */
		for (pp = ptbl; ; pp++) {
			if (pp->p_id == '\0') {
				logit( "NO SUPPORTED PROTOCOLS", protocols );
				return FAIL;
			}
			if (strchr (protocols, pp->p_id) != NULL)
				break;
		}
	}

	/* To debug dialout, login, etc in ASCII, print easy debug logs */
	Debug_data = debug_data_easy;

	port = (struct port *)0;
	for (;;) {
		port = pickport(acu, baud, port);
		if (!port) {
			DEBUG(2, "Failed opening all ports...\n", 0);
			return FAIL;		/* Tried them all */
		}
		DEBUG(2, "Opening outgoing line %s\n", port->devname);
		if (openout(port->devname, baud) != SUCCESS) {
			/* It would be good to log the reason, e.g.
			 * lockfile in use, or port wouldn't open.  FIXME. */
			logit("CAN'T USE", port->devname);
		} else {
			break;
		}
	}

	sprintf(logbuf, "%s %s %ld", host_name, port->devname, port->baud);
	DEBUG(1, "Trying %s\n", logbuf);

	if (dial_nbr(port, telno)) {
		hangup();
		logit("FAILED", logbuf);
		return FAIL;
	}

	logit("DIALED", logbuf);

	/*
	 * Process send-expect strings.
	 */
	while (expect = strtok((char *)NULL, " ")) {
		send = strtok((char *)NULL, " ");
		if (sendexpect(expect, send) != SUCCESS)
			goto bort1;
	}

	/*
	 * FIXME, there should be a way to detect login/passwd
	 * failure here and keep doing the script rather than
	 * continuing to expect Shere at another login: prompt.
	 */
	sprintf(logbuf, "call to %s", host_name);
	logit("SUCCEEDED", logbuf);

	/* wait for Shere message, grab whatever is sent, send response */
	sprintf(tempname, "\20Shere");
	if (instr(tempname, strlen(tempname)))
		goto bort1;
	if (getstring(msgbuf) != SUCCESS)
		goto bort1;
	if (msgbuf[0] == '=') {
		if (!!strcmp(host_name, msgbuf+1)) {
			logit("WRONG HOST", msgbuf);
			goto bort1;
		}
	} else if (msgbuf[0] != '\0') {
		DEBUG(0, "strange Shere%s\n", msgbuf);
	}

	sprintf(tempname, "\20S%s\0", Myname);
	/* For some reason, the Amiga version had:
	 *   sprintf(tempname, "\20S%s -Q0 -x%d\0", Myname, debug_level);
	 * but I am not sure what -Q0 does, or why we want the other
	 * end to know that we are debugging this end.   --gnu@toad.com
	 * Phil Budne's version had a comment about passing -g -xN -Qn -pC
	 * and -vgrade=C.  I STILL don't know what any of these mean. */
	twrite(tempname, strlen(tempname)+1); /* Including null */

	/* wait for ok message */
	if (getstring(msgbuf) != SUCCESS)
		goto bort1;
	if (msgbuf[0] != 'R')
		goto bort1;
	if (strcmp("OK", msgbuf+1)) {
		logit("HANDSHAKE FAILED", msgbuf+1);
		goto bort1;
	}

	/*
	 * Get Protocol string, make sure it supports our protocol.
	 * FIXME, should scan our table versus theirs.
	 */
	if (getstring(msgbuf) != SUCCESS)
		goto bort1;
	if (msgbuf[0] != 'P')
		goto bort1;

	/* scan our table for first match.. (honor OUR ordering!) */
	/* perhaps use oridering from user (protocols) */
	for (pp = ptbl; ; pp++) {
		if( (protocols == NULL	/* no restrictions */
		     || strchr (protocols, pp->p_id) != NULL) /* or match */
		    && strchr (msgbuf+1, pp->p_id) != NULL) /* and offered */
			break;			/* go for it! */
		if (pp->p_id == '\0') {
			twrite(msg_noproto, sizeof(msg_noproto)-1);
			logit("NO PROTOCOL IN COMMON", msgbuf);
			goto bort1;
		}
	}

	/*
	 * Tell the other side which protocol to use.
	 */
	sprintf(tempname, "\20U%c", pp->p_id);
	twrite(tempname, strlen(tempname)+1);	/* Including null */

	curproto = pp;		/* Tell our side, too... */

	/* Now that we aren't running in ASCII, print detailed debug logs */
	Debug_data = debug_data_ugly;

	if( turnon(1) != SUCCESS ) {
		sprintf(logbuf, "call to %s", host_name);
		logit("PROTOCOL TURNON FAILED", logbuf);
		goto bort1;
	}

	logit("OK", "startup");

	top_level(1);
	hangup();
	return SUCCESS;

bort1:
	hangup();
	return FAIL;
}

/* Handle a single uucp [slave] login session */
void
do_session(ontheline)
	int ontheline;
{
	char	trash[MAX_STRING+SLOP];	/* trash I/O buffer */
	register char *tp;
	register struct proto *pp;

	if (!ontheline) {
		/* output login request, verify uucp */
		twrite(msgo0,sizeof(msgo0)-1);
		if (instr(msgi0,sizeof(msgi0)-1)) {
			printf("gnuuio: invalid login name\n");
			goto bort;
		}

		/* output password request, verify s8000 */
		twrite(msgo1,sizeof(msgo1)-1);
		if (instr(msgi1,sizeof(msgi1)-1)) {
			printf("gnuuio: invalid password\n");
			goto bort;
		}

		printf("gnuuio: correct login\n");
	}

	/* output here message, wait for response */
	sprintf(trash, "\20Shere=%s", Myname);
	twrite(trash, strlen(trash)+1);	/* Including null */

	if (getstring(trash) != SUCCESS || trash[0] != 'S')
		goto bort;
	strcpy (host_name, strtok(trash+1, " "));
	/* FIXME, parse flags that follow hostname too */

	/* output ROK message */
	twrite(msgo3,sizeof(msgo3)-1);

	/* List all supported protocols, and output that as a "P" message */
	tp = trash;
	*tp++ = '\20';
	*tp++ = 'P';
	for (pp = ptbl; pp->p_id; pp++)
		*tp++ = pp->p_id;
	*tp++ = '\0';
	twrite(trash, tp-trash);
	
	/* Check protocol response */
	if (getstring(trash) != SUCCESS || trash[0] != 'U')
		goto bort;
	for (pp = ptbl; ; pp++) {
		if (pp->p_id == '\0') {
			logit("UNKNOWN PROTOCOL", trash+1);
			goto bort;
		}
		if (pp->p_id == trash[1]) 
			break;
	}
			
	curproto = pp;		/* Enable use of selected protocol */
	if (turnon(0) != SUCCESS) {
		logit("FAILED PROTOCOL STARTUP", trash+1);
		goto bort;
	}
	logit("OK", "startup");
	top_level(0);		/* Returns SUCCESS or FAIL, we dont care */

bort:
	hangup();	/* FIXME, what *exactly* does this do? -- gnu */
	printf("gnuuio: call complete\n");
}

/*
 * Handle transactions "at top level", as Unix uucp's debug log says.
 *
 * As master, we scan our queues for work and send requests to the
 * other side.  When done, we send a hangup request and switch to slave mode.
 *
 * As slave, we accept requests from the other side; when it is done,
 * it sends a hangup request, and we switch to master mode, if we have
 * any work queued up for that system.
 *
 * This repeats as long as either side has work to do.  When all the
 * queued work is done, we agree to hang up, terminate the packet protocol,
 * and return to the caller.  (We still haven't hung up the phone line yet.)
 *
 * A curious feature of the hangup protocol is that it is not a simple
 * question-answer.  The master says "H", asking about hangup.  The
 * slave responds "HY" saying OK.  The master then says "HY" also,
 * then both of them hang up.  Maybe this is to make sure the first HY
 * got ack'ed?  Anyway, an "H" is reported as HANGUP and an "HY" as
 * HANGNOW.  After we send an HY, we go back to listening for commands;
 * if the master sends something other than HY, we'll do it.
 */
s_or_f
top_level(master_mode)
	int master_mode;
{
	char	buf[MAXMSGLEN];		/* For hangup responses */

	if (master_mode) {
		(void) work_scan(host_name, "C"); /* Kick off queue scan */
		goto master;
	}

	for (;;) {
		/* Slave side */
	slave:
		for (;;) {
			DEBUG(4, "*** TOP *** - slave\n", 0);
			switch (do_one_slave()) {
			case WSUCCESS:
			case COPYFAIL:	/* FIXME, consider other action */
				break;
			case WFAIL:
				goto failing;
			case HANGUP:
				if (work_scan(host_name, "C")) {
					if (wrmsg('H', "N", fdtty) != SUCCESS)
						goto failing;
					goto master;
				} else {
					if (wrmsg('H', "Y", fdtty) != SUCCESS)
						goto failing;
					break;
				}
			case HANGNOW:
				goto quit;
			}
		}
	
		/* Master side */
	master:
		for (;;) {
			DEBUG(4, "*** TOP *** - master\n", 0);
			switch (do_one_master()) {
			case WSUCCESS:
			case COPYFAIL:	/* FIXME, consider other action */
				break;
			case WFAIL:
				goto failing;
			case HANGUP:
				/* We wrote an H command, what's the resp? */
				if (rdmsg(buf, fdtty) != SUCCESS)
					goto failing;
				if (buf[0] != 'H')
					goto failing;
				if (buf[1] == 'N') {
					DEBUG(4,"Slave accepts turnaround\n",0);
					goto slave;
				} else {
			case HANGNOW:	/* Should not happen, but... */
					/* Write the final HY */
					DEBUG(4, "Hanging up\n", 0);
					if (wrmsg('H', "Y", fdtty) != SUCCESS)
						goto failing;
					goto quit;
				}
			}
		}
	}

quit:
	/* Shut down the packet protocol */
	turnoff();

	/* Write the closing sequence.  If we don't get it in response,
	   hang up the phone anyway -- don't waste time and $ arguing
	   about whether to hang up at this point! */
	twrite(msgo4, sizeof(msgo4)-1);
	(void) instr(msgi4, sizeof(msgi4)-1);

	twrite(msgo4, sizeof(msgo4)-1);	/* Send again to help other side */

	logit("OK", "conversation complete");

	return SUCCESS;   /* Go byebye */

failing:
	logit("FAILED", "conversation complete");
	return FAIL;
}

/*
 * We are slave; get a command from the other side and execute it.
 */
line_status
do_one_slave()
{
	char msg[MAXMSGLEN];		/* Master's message to us */

	/* Get master's command */
	if (rdmsg(msg, fdtty) != SUCCESS)
		return WFAIL;

	/* Print it for easy debugging */
	DEBUG(4,"Command: %s\n\n", msg);

	switch (msg[0]) {

	case 'S':
		if (msg[1] != ' ') break;
		return host_send_file(msg);

	case 'R':
		if (msg[1] != ' ') break;
		return host_receive_file(msg);

	case 'X':
		/* Do proxy execution of uucp (expand *, etc, on this side);
		 * X src destm!destf user flags
		 */
		break;

	case 'H':
		if (msg[1] == '\0') return HANGUP;
		if (msg[1] == 'Y')  return HANGNOW;
		if (msg[1] == 'N')  return WSUCCESS;	/* Ignore HN to slave */
		break;

	}

	/* Unrecognized packet from the other end */
	DEBUG(0, "Bad control packet refused: %s\n", msg);
	/* Overwrite it with the error response. */
	msg[1] = 'N'; msg[2] = '\0';		/* FIXME: return error code */
	if (yesno(msg))
		return WFAIL;
	return WSUCCESS;
}

/*
 * Do one piece of work as master.
 *
 * FIXME:  we don't handle the flags, e.g. -c, properly!
 */
line_status
do_one_master()
{
	FILE	*fd;
	char	*sname;
	char	cmnd[1];		/* Command character */
	char 	buf[100];
	line_status	fail;
	int	hang = 0;
	int	num;
	char	notify[NAMESIZE];	/* A bit large...FIXME */
			/* FIXME: do the notify stuff */
	static char	logseq[WSUFSIZE+1];	/* For logging seq # */

	sname = work_next();
	if (!sname) {
		/* No more work, time to hang up. */
		DEBUG(4, "No more work.\n", 0);
	hangitup:
		DEBUG(4, "Requesting line turnaround\n", 0);
		if (wrmsg('H', "", fdtty))
			return WFAIL;
		return HANGUP;
	}

	DEBUG(4, "Request file %s\n", sname);

	fd = fopen(munge_filename(sname), "r");
	if (fd == NULL) {
		DEBUG(0, "gnuuio: couldn't open %s\n", sname);
		/* Some other uucp may have gotten it */
		return WSUCCESS;
	}

	/* Put the C. file sequence number in future log entries */
	strcpy(logseq, sname+strlen(sname)-WSUFSIZE+1);	/* Copy seq # */
	log_sequence = logseq;

	while (fgets(buf, sizeof buf, fd)) {
		DEBUG(4, "Queued request: %s", buf);
		if (buf[1] != ' ') goto badnum;
		num = sscanf(buf, "%s %s %s %s %s %s %o %s\n",
			cmnd, srcnam, dstnam, who, flags, temp, &mode, notify);

		if (num < 5)
			flags[0] = '\0';

		switch (cmnd[0]) {
		case 'S':
			if (num < 7 || num > 8) goto badnum;
			fail = local_send_file();
			/* Handle remote-requested turnaround -- we might
			   have some more files to do first. */
			if (fail == HANGUP) {
				DEBUG(4, "Line turnaround requested\n", 0);
				hang++;
				fail = WSUCCESS;
			}
			break;

		case 'R':
			if (num != 5) goto badnum;
			fail = local_receive_file();
			break;

		case 'X':
			if (num < 3 || num > 5) goto badnum;
			/* FIXME */
			/* fall thru to default */

		default:
		badnum:
			DEBUG(0, "Unknown/invalid queued request: %s\n",
				 buf);
			goto badline;
		}

		/* FIXME, what does uucp do if one of N xfers fails? */
		if (fail == WFAIL) {
			/* protocol over the wire is failing */
			logit("PROTOCOL PROBLEM", "");
			fclose(fd);
			log_sequence = 0;	/* Stop C. file logging */
			return WFAIL;
		}
		if (fail != WSUCCESS) {
	badline:
			logit("ERROR IN WORK FILE", sname);
			buf[strlen(buf)-1] = '\0';   /* Zap newline */
			logit("BAD LINE IS", buf);
		}
	}
	fclose(fd);

	log_sequence = 0;		/* Stop logging this C. file */

	/* Zap the queue file */
	/* FIXME, We should consider what to do here when a COPYFAIL occurs. */
	if (remove(munge_filename(sname)) != 0) {
		logit("CAN'T REMOVE WORK FILE", sname);
		DEBUG(0, "Can't remove, errno %d\n", errno);
	} else {
		DEBUG(4, "Removed work file %s\n", sname);
	}

	if (hang)
		goto hangitup;
	return WSUCCESS;
}

/* Send a "yes or no" packet with character 'c'. */
s_or_f
yesno(string)
	char *string;
{

	return wrmsg(string[0], string+1, fdtty);
}

/*
 * Master wishes to send a file to us -- we receive it.
 * Return 1 to abort the call, 0 to continue.
 */
line_status
host_send_file(msg)
char  *msg;
{
	FILE *fddsk;			/* Disk file pointer */
	char	cmnd[1];		/* Command character */

	sscanf(msg,"%s %s %s %s %s %s %o",
		cmnd, srcnam, dstnam, who, flags, temp, &mode);
	logit("REQUESTED", msg);
	strcpy (dstnam, munge_filename(dstnam));  /* Translate to local name */
	strcpy (temp, temp_filename(dstnam));     /* Create a handy temp file */

	fddsk = fopen(temp, "wb");
	if (fddsk == NULL) {
		/* Can't open file -- send error response */
		if (DEBUG_LEVEL(0)) {
		    printf(
			"Cannot open temp file %s (%s) for writing, errno=%d\n",
				temp, dstnam, errno);
		}
		logit("REQUEST", "FAILED -- TEMP FILE");
		if (yesno("SN4"))
			return WFAIL;
		return WSUCCESS;
	}

	/* setmode (fddsk, temp, mode|PROT_RW_OWNER);	Set file modes */
	/* FIXME: Should we create directories for the file? */
	if (yesno("SY") != SUCCESS)    /* Say yes */
		return WFAIL;

	return receive_file(fddsk, temp, dstnam);
}

/*
 * Master wants to Recieve a file from us -- we send it.
 * Return 1 to abort the call, 0 to continue.
 */
line_status
host_receive_file(msg)
char  *msg;
{
	FILE *fddsk;     /* Disk file descriptor */
	line_status x;
	char	cmnd[1];		/* Command character */
	char msg2[MAXMSGLEN];

	logit("REQUESTED", msg);

	sscanf(msg,"%s %s %s",cmnd,srcnam,dstnam);
	strcpy (temp, munge_filename(srcnam));

	fddsk = fopen(temp, "rb");		/* Try to open the file */
	if (fddsk == NULL) {
		/* File didn't open, sigh. */
		if (DEBUG_LEVEL(0)) {
		    printf("Cannot open file %s (%s) for reading, errno=%d\n",
				temp, srcnam, errno);
		}
		logit("DENIED", "CAN'T OPEN");
		if (yesno("RN2") != SUCCESS)
			return WFAIL;
		return WSUCCESS;
	}

	/* FIXME, we should be figuring out the file's permission bits
	 * here and sending them to the other side.  For now, we just send
	 * "0666" which is read and write for everybody.  Fixing this
	 * requires sysdep routines to have "get file permission" and
	 * "set file permission" routines, which tonight I'm too lazy to do.
	 */
	sprintf(msg2, "RY 0%o", 0666 /* getmode(fddsk, temp)*/ );
	if (yesno(msg2) != SUCCESS)    /* Say yes */
		return WFAIL;

	x = send_file(fddsk);
	switch (x) {
	default:
		return x;

	case COPYFAIL:
		/* We don't care if the copy failed, since the master
		   asked for the file and knows the result. */
		return WSUCCESS;
	}
}

/*
 * We, as master, want to Send a file.
 *
 * Return WFAIL, WSUCCESS, HANGUP, or COPYFAIL.
 * SUCCESS is returned either if the file was not found locally (local
 * error, and the queued transfer should be flushed) or if it was moved
 * successfully.  COPYFAIL indicates that the queued transfer should be
 * left queued, and later retried.  HANGUP indicates that the other
 * side wants to gain control and we should offer to hang up.  WFAIL
 * indicates that a protocol failure has occurred and we should give up.
 * FIXME, there are several failure points
 * in the transaction (see Protocol.doc) and we need finer control here.
 */
line_status
local_send_file()
{
	char buf[MAXMSGLEN];	/* Used for both xmit and receive */
	FILE *fddsk;		/* Disk file descriptor */
	line_status res;	/* Result status */
	int status;		/* File removal status */

	/* FIXME, notify if requested. */
	sprintf(buf,"S %s %s %s %s %s 0%o %s",
		srcnam, dstnam, who, flags, temp, mode, who);

	logit("REQUEST", buf);

	if (strchr(flags, 'c')) {
		strcpy(temp, munge_filename(srcnam));
	} else {
		strcpy(temp, munge_filename(temp));
	}
	fddsk = fopen(temp, "rb");
	if (fddsk == NULL) {
		/* FIXME -- handle queued request for nonexistent file */
		if (DEBUG_LEVEL(0))
			printf("Can't open file %s (%s), errno=%d\n",
				temp, srcnam, errno);
		logit("NOT FOUND", temp);
		return COPYFAIL;	/* FIXME caller won't deal with this */
	}

	/* Tell the other side we want to send this file */
	if (wrmsg('S', buf+1, fdtty) != SUCCESS) {
		DEBUG(0, "problem sending request\n", 0);
		return WFAIL;
	}

	/* See what they have to say about it */
	if (rdmsg(buf, fdtty) != SUCCESS)	
		return WFAIL;
	if ((buf[0] != 'S') || (buf[1] != 'Y')) {
		logit("REQUEST DENIED", buf);
		return COPYFAIL;
	}
	res = send_file(fddsk);	/* WFAIL, WSUCCESS, or COPYFAIL */

	/* Delete the source file if it was successful and just a copy */
	if (res != WSUCCESS && res != HANGUP)
		return res;
	if (strchr(flags, 'c'))		/* If copied direct from source */
		return res;		/* ...just return. */
	status = remove(temp);		/* Delete uucp's copy of the file */
	if (status != 0) {
		logit("CAN'T REMOVE SENT FILE", temp);
		DEBUG(0, "Can't remove, errno %d\n", errno);
	} else {
		DEBUG(4, "Removed sent file %s\n", temp);
	}
	return res;
}

/*
 * We want to Receive a file -- so we ask for it.
 * Return 1 to abort the call, 0 to continue.
 */
line_status
local_receive_file()
{
	char buf[MAXMSGLEN];
	FILE *fddsk;			/* Disk file pointer */

	/* FIXME, test dest file access before we ask for it. */

	sprintf(buf,"R %s %s %s %s %s 0%o %s",
		srcnam, dstnam, who, flags, temp, mode, who);

	strcpy (dstnam, munge_filename(dstnam));  /* Translate to local name */
	strcpy (temp, temp_filename(dstnam));     /* Create a handy temp file */

	/* FIXME: deal with file modes now that we fopen. */
	/* FIXME: Are the above permissions right?? */
	/* FIXME: Should we create directories for the file? */
	fddsk = fopen(temp, "wb" /*, mode|060 */);

	if (fddsk == NULL) {
		/* Can't open temp file -- send error response */
		if (DEBUG_LEVEL(0)) {
		    printf(
			"Cannot open temp file %s (%s) for writing, errno=%d\n",
				temp, dstnam, errno);
		}
		logit("REQUEST", "FAILED -- TEMPFILE");
		return WSUCCESS;
	}

	logit("REQUEST", buf);
	if (wrmsg('R', buf+1, fdtty) != SUCCESS) {
		printf("gnuuio: problem sending request\n");
		return WFAIL;
	}

	/* See what the other side has to say about it */
	if (rdmsg(buf, fdtty) != SUCCESS)
		return WFAIL;
	if ((buf[0] != 'R') || (buf[1] != 'Y')) {
		logit("REQUEST DENIED", buf);
		return WSUCCESS;	/* FIXME, should do something more here */
	}

	return receive_file(fddsk, temp, dstnam);
	/* FIXME - We should deal with files that didn't get there */
	/* FIXME -- setmode */
}

/* general file receive routine
 *
 * The "temp" file has been created and opened with fddsk.  Dstnam is the
 * eventual file name we want this file to have.
 */
line_status
receive_file(fddsk, temp, dstnam)
	FILE *fddsk;			/* Disk file pointer */
	char	*temp, *dstnam;
{
	int status;
	int error = 0;			/* No errors so far */

	if (rddata(fdtty, fddsk) != SUCCESS) error++;
	status = fclose(fddsk);		/* Make sure the data got here */
	if (status != 0) {
		error++;
		DEBUG(0, "fclose errno=%d\n", errno);
	}

	/* Move the file from its temp location to its real loc */
	/* FIXME:  This needs to be able to copy the file, if 
	   a simple rename does not suffice. */
	/* FIXME:  should create directories if necessary, e.g. D.
	   or subdirs of /usr/spool/uucppublic. */
	/* FIXME:  should use source name if target is a directory e.g. ~/ */
	status = rename(temp, dstnam);
	if (status != 0) {
		error++;
		if (DEBUG_LEVEL(0)) {
			printf("Cannot rename file %s to %s, errno=%d\n",
				temp, dstnam, errno);
		}
	}

	logit("COPY", error? "FAILED": "SUCCEEDED");
	/* FIXME: if slave, and want a chance send "CYM" rather than "CY". */
	if (yesno(error == 0? "CY": "CN5"))	/* Send yes or no */
		return WFAIL;
	return WSUCCESS;
}

/*
 * general file send routine
 * Return SUCCESS, FAIL, or COPYFAIL.
 */
line_status
send_file(fddsk)
	FILE *fddsk;     /* Disk file pointer */
{
	char ansbuf[MAXMSGLEN];

	if (wrdata(fddsk, fdtty) != SUCCESS)
		return WFAIL;
	(void) fclose(fddsk);
	/* Await the "CY" or "CNddd" packet, and toss it. */
	while (1) {
		if (rdmsg(ansbuf, fdtty) != SUCCESS)	
			return WFAIL;
		if (ansbuf[0] != 'C') {
			DEBUG(0,"\nDidn't get 'CY' or 'CN', got %s\n",
				ansbuf);
			/* and loop looking for C message */
		} else if (ansbuf[1] == 'Y') {
			logit("REQUESTED", ansbuf);
			if (ansbuf[2] == 'M') {
				/* We should let other side be master */
				return HANGUP;	
			}
			return WSUCCESS;
		} else {
			logit("COPY FAILED", ansbuf);
			return COPYFAIL;
		}
	}
}

/* 
 * See if the current time is within the string "HHMM-HHMM" ala 
 * the l.sys file.  Times are in 24 hour format.  The string can 
 * cross midnight (ie. 2300-0800).
 *
 * Returns 1 if the time is within the window, 0 if not.
 */
int
istime(time_str)
	char *time_str;
{
	int	b_time,		/* ok to call begging from this time... */
		e_time,		/* to this end time from l.sys */
		l_time;		/* current local time */
	struct tm *nt;
	time_t	long_time;
	char	*junk;

	time(&long_time);
	nt = localtime(&long_time);
	l_time = (nt->tm_hour * 100) + nt->tm_min;

	if (2 != sscanf(time_str, "%d-%d%s", &b_time, &e_time, &junk)) 
		return 0;
	
	if (b_time > 2400)
		DEBUG(0, "%d is not a valid start time to call.", b_time);
	if (e_time > 2400)
		DEBUG(0, "%d is not a valid end time to call.", b_time);

	if (b_time < e_time) {		/* e.g. 0800-1700 */
		if ((l_time >= b_time) && (l_time <= e_time)) {
			return 1;
		}
	} else {			/* e.g. 2300-0730 */
		if ((l_time >= b_time) || (l_time <= e_time)) {
			return 1;
		}
	}
	return 0;
}

/*
 * Return 1 if right now is a valid time according to the string.
 * Return 0 if not a good time, or invalid string.
 */
int 
is_time_ok(str)
	register char *str;
{

	register char *word;
	char wordbuf[30];		/* For lowercasing */
	register struct timetable *tt;
	register char *low;
	struct tm *nt;
	time_t	long_time = 0;
	
	while (*str) {

		word = str;
		while (isalpha(*str)) str++;
		
		if (str > word) {
			/* We found a word indicating a day-of-week. */
			if (str - word > sizeof(wordbuf)-1)
				goto failday;	/* Unknown word */
			for (low = wordbuf; word < str; word++, low++) {
				if (isupper(*word))
					*low = tolower(*word);
				else
					*low = *word;
			}
			*low = '\0';		/* Null terminate */

			for (tt = Timetable; tt; tt = tt->next) {
				if (strcmp(tt->str, wordbuf) == 0) {
					/* Found the word in table.  Do it. */
					switch (tt->what) {
					case Day:
						if (long_time == 0) {
							/* Fetch cur time */
							time(&long_time);
							nt = localtime
							       (&long_time);
						}
						if (tt->dayval != nt->tm_wday)
							goto failday;
						goto succeedday;

					case Any:
						goto succeedday;

					case Recurse:
						if (is_time_ok(tt->recurseval))
							goto succeedday;
						else
							goto failday;

					default:
						abort();
					}
				}
			}
			goto failday;		/* Unknown word */
		
		}

		/*
		 * The day-of-week part of the spec has succeeded, or was
		 * not present.  Check for a time spec.
		 */
	succeedday:
		word = str;
		while (isdigit(*str) || *str == '-') str++;
		if (str > word) {	/* if there is time spec'd */
			if (word - str > sizeof(wordbuf)-1)
				goto failtime;		/* Too long */
			strncpy(wordbuf, word, str-word);
			wordbuf[str-word] = '\0';
			if (!istime(wordbuf)) {
				goto failtime;
			}
		}

		/*
		 * We have passed both day and time checks.  See if
		 * valid terminating character.
		 */
		if (*str == '\0' ||
		    *str == '|' ||
		    *str == ',' ||
		    *str == ';' )
			return 1;		/* Valid */

		/*
		 * This part of the spec is not matching now.  Skip
		 * forward until the next part.
		 */
	failday:
	failtime:
		while (*str != '\0' &&
		       *str != '|' &&
		       *str != ',' &&
		       *str != ';' )
			   str++;
	}

	return 0;			/* No part matched now. */
}
