/*
    $Header: uux.c,v 1.6 90/04/25 23:19:00 gnu Exp $
    Copyright 1990 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    uux - queue up a job to execute on another system (or this system)

    Originally written by <WSWIETSE@HEITUE5.BITNET>
 */

#ifndef lint
static char rcsid[] = "$Header: uux.c,v 1.6 90/04/25 23:19:00 gnu Exp $";
#endif

 /*
  * uux [-gGRADE] [-n] [-z] [-xDEBUGLEVEL] [-] [host!]command [arguments]
  * 
  * -gX sets the job "grade"; this allows you to give different priorities to
  *    news and mail.
  * -n suppresses all reports about completion (success or failure).
  * -r do not start uuxqt (local command) or uucico (remote command).
  * -xN sets the debugging level.
  * -z suppresses reports about successful completion (still get failure notes).
  * -, or -p, requests that standard input be copied as standard input to the
  *    command when it is executed.
  * 
  * Exit status codes are in agreement with sendmail(8).
  *
  * The present version does not implement host!file arguments, but understands
  * the ( ) quoting convention needed by mail.
  * 
  * We do not yet start execution of uuxqt (local command) or uucico (remote
  * command).
  * 
  * Need to clean up files if we terminate with an error. Need to set up
  * signal handlers.
  */

#include "includes.h"
#include "uucp.h"
#include "sysdep.h"

/* Define the following values if you do not have <sysexits.h> */

#define	EX_USAGE	64
#define	EX_NOHOST	68
#define	EX_CANTCREAT	73

 /*
  * Things to be done. See also the TODO: comments throughout the source.
  * 
  * The following macros need suitable definitions for the final uux version.
  * 
  * uucp_cmd() should run a command that lives in the local equivalent of
  * /usr/lib/uucp.
  */

/* More stuff that will have to find its place in sysdep.[hc] */

#ifdef unix
#define rename(from,to)	(link(from,to) || unlink(from))
#endif

extern char *getenv();
extern void exit();

extern s_or_f read_params();
extern char *getseqnum();
extern void logit();

extern int errno;

 /*
  * Global variables
  */

static char grade = DEF_GRADE;		/* grade of job */
static char nflag = 0;			/* do/don\'t report success */
static char rflag = 0;			/* do/don\'t start up uucico */
static char zflag = 0;			/* do/don\'t report errors */
static char do_stdin = 0;		/* do/don\'t copy standard input */
static char *xcmd;			/* command to be executed */
static int xlast;			/* first free position in xcmd */
static unsigned xcmdsize;		/* malloc'd size of xcmd */
static int remote_work;			/* true if execution will be remote */

/* The following are set by bang_split() */

static char *hostpart;			/* host part of an ordinary argument */
static char *otherpart;			/* non-host part of ordinary argument */

/* Characters that receive special treatment, to some extent... */

static char specials[] = "'`\"(<>";	/* special at beginning of argument */
static char starts[] = "'`\"(";		/* these start a special argument */
static char stops[] = "'`\")";		/* these terminate a special argument */

/* Forward declarations */

static void parse_cmd();
static void bang_split();
static char *upto();
static char *getarg();
static void build_cmd();
static void uucpname();
static FILE *spoolcreate();
static void spoolmove();
static void write_files();
static void error();
static void usage();

/* I believe main() is the only global symbol in this file. */

main(argc, argv)
int     argc;
char  **argv;
{
    register char	*buf, *bp;		/* Command line buffer */
    register unsigned	bufsize;		/* Its size */
    register int	i;
    char 		*username;

    /* Parse command-line options */

    while (--argc && *++argv && **argv == '-') {
	switch (*++*argv) {
	case 'g':				/* non-default job grade */
	    grade = (*++*argv ? **argv : DEF_GRADE);
	    break;
	case 'n':				/* don\'t report success */
	    nflag = 1;
	    break;
	case 'r':				/* don\'t start uucico */
	    rflag = 1;
	    break;
	case 'x':				/* debugging level */
	    debug = atoi(*argv + 1);
	    break;
	case 'z':				/* don\'t report errors */
	    zflag = 1;
	    break;
	case 'p':				/* copy standard input */
	case '\0':				/* copy standard input */
	    do_stdin = 1;
	    break;
	default:
	    usage();
	    /* NOTREACHED */
	}
    }

    if (argc == 0)
	usage();

    if (read_params((char *)0) != SUCCESS)
	exit(1);

    username = getenv("USER");
    if (username == 0 || username[0] == '\0')
	username="uucp";

    strcpy(who, username);		/* For logging and such */

    /*
     * Concatenate all arguments into one big string.
     */
    bufsize = 0;
    for (i = 0; i < argc; i++)			/* Count size needed */
	bufsize += 1 + strlen(argv[i]);

    buf = malloc(bufsize + SLOP);		/* Use that much */

    for (bp = buf; argc--; argv++) {
	(void) strcpy(bp, *argv);		/* Copy in one arg */
	bp += strlen(*argv);
	*bp++ = ' ';				/* Add space after */
    }
    bp[-1] = '\0';				/* Delete last space */

    /*
     * Parse remainder of command arguments. As long as we disallow file
     * arguments, the writing of spool files can be postponed until the uux
     * command has been fully parsed.
     */

    parse_cmd(buf, bufsize);

    /*
     * Make spool files. As long as we disallow file arguments we can write
     * spool files after we have fully parsed the uux command.
     */

    write_files();

    /*
     * Start command execution. The uuxqt command on the execution host will
     * check whether the command is allowed or not, and, if it is smart
     * enough, will mail the command completion status back to the user.
     */

    if (rflag == 0) {
	if (remote_work) {			/* non-local command */
	    if (Gnuuio)
		execlp(Gnuuio, "-gnuuio", "-s", host_name, (char *) 0);
	} else {				/* local command */
	    if (Gnuuxqt)
		execlp(Gnuuxqt, "-gnuuxqt", (char *) 0);
	}
    }

    exit(0);
    /* FIXME: for some bloody reason, MSC 5.1 is giving a run-time error of:
              "run-time error R6001 - null pointer assignment" 
              and exiting with a '-1' at this exit() */
    /* NOTREACHED */
}

int
knownhost(h)
	char *h;
{
	FILE	*lsys;
	char	buf[MAX_LSYS];
	char	sysnam[MAX_HOST];

	if (! (lsys = fopen(Sysfile, "r"))) {
		DEBUG(0, "knownhost: can't open %s", Sysfile);
		DEBUG(0, ", errno %d\n", errno);
		return 0;			/* Didn't even find L.sys */
	}
	/* Once per system in L.sys... */
	/* FIXME, handle continuation lines (trailing "\") */
	while (fgets(buf, sizeof buf, lsys)) {
		if (buf[0] == '#')
			continue;
		
		/* Grab the system name. */
		(void) sscanf(buf, "%s", sysnam);

		if (strcmp(h, sysnam) == SAME) {
			fclose(lsys);
			return 1;		/* Found H in L.sys */
		}
	}

	fclose(lsys);
	return 0;			/* Didn't find H in L.sys */
}


/* parse_cmd - establish execution host, command and arguments */

static void parse_cmd(arg, argsize)
    char   *arg;
    unsigned argsize;
{
    char   *argp = arg;
    char   *getarg();
    char    *argval = malloc(argsize+SLOP);	/* Won't be bigger than this */

    xcmdsize = argsize;			/* Start off with same space */
    xcmd = malloc(argsize + SLOP);	/* But build_cmd can realloc it */
    xlast = 0;

    if (argval == 0 || xcmd == 0)
	error (EX_CANTCREAT, "Out of memory in malloc()", "");

    while (argp = getarg(argp, argval)) {
	hostpart = otherpart = 0;		/* build_cmd() needs this */
	if (strchr(specials, argval[0]) == 0) {
	    bang_split(argval);
	    if (xlast > 0 && hostpart) {		/* file argument */
		error(EX_USAGE, "host!file is not supported", "");
	    } else {				/* non-file argument */
		build_cmd(otherpart);
	    }
	} else if (argval[0] == '>' || argval[0] == '<') {
	    error(EX_USAGE, "i/o redirection is not supported", "");
	} else if (argval[0] == '(') {		/* ( ) quoted argument */
	    argval[strlen(argval) - 1] = '\0';
	    build_cmd(argval + 1);
	} else {				/* anything else */
	    build_cmd(argval);
	}
    }
    xcmd[--xlast] = '\0';		/* Zap trailing blank */
}

/* bang_split - split argument on bang character, expand null host names */

static void bang_split(arg)
register char *arg;
{
    register char *cp;

    if ((cp = strchr(arg, '!')) == 0) {		/* stuff */
	hostpart = 0;
	otherpart = arg;
    } else {					/* [host]!stuff */
	*cp++ = 0;
	if (arg[0] == 0)			/* !stuff */
	    hostpart = Myname;
	else if (knownhost(arg))		/* known-host!stuff */
	    hostpart = arg;
	else					/* unknown-host!stuff */
	    error(EX_NOHOST, "unknown host: %s", arg);
	otherpart = cp;
    }
    DEBUG(9, "hostpart:  %s\n", hostpart ? hostpart : "(none)");
    DEBUG(9, "otherpart: %s\n", otherpart);
}

/* upto - copy up to and including terminator; always copy leading character */

static char *upto(from, to, term)
char   *from;
char   *to;
int     term;
{
    char   *cp;

    if ((cp = strchr(from + 1, term) + 1) == (char *) 1)
	error(EX_USAGE, "missing %c character", (char *)term);

    while (from < cp)
	*to++ = *from++;
    *to = '\0';
    return (cp);
}

/* getarg - extract next argument, return next search position */

static char *getarg(from, to)
char   *from;
char   *to;
{
    char   *sp;

    /* Skip leading blanks */

    while (isspace(*from))
	from++;
    if (*from == '\0')
	return (0);

    /* Handle various forms of quoting */

    if (sp = strchr(starts, *from))
	return (upto(from, to, stops[sp - starts]));

    /* Copy up to next white space or end of string */

    while (*from && !isspace(*from))
	*to++ = *from++;
    *to = '\0';
    return (from);
}

/* build_cmd - append next argument to command, determine execution host */

static void build_cmd(arg)
char   *arg;
{
    int     arglen = strlen(arg);

    /* Special case if we are copying the first part of a command */

    if (xlast == 0) {
	(void) strcpy(host_name, hostpart ? hostpart : Myname);
	DEBUG(4, "host_name: %s\n", host_name);
	remote_work = strcmp(host_name, Myname);
    }

    /* Make sure that we don't run out of buffer space with long commands */

    if (xlast + arglen + 1 >= xcmdsize) {
	xcmd = (char *)realloc(xcmd, xcmdsize*2);
	xcmdsize *= 2;
    }

    (void) strcpy(xcmd+xlast, arg);
    xlast += arglen;
    xcmd[xlast++] = ' ';		/* Space after arg */
    xcmd[xlast]   = '\0';		/* Terminate string */

    DEBUG(4, "append: %s\n", arg);
}

/* uucpname - build canonical uucp file name */

static void uucpname(buf, type, host, gchar, seqno)
char   *buf;
int     type;
char   *host;
int     gchar;
char   *seqno;
{
    (void) sprintf(buf, "%c.%.*s%c%s", type, MAXBASENAME, host, gchar, seqno);
    return;
}

/* spoolcreate - map canonical uucp file name to local spool file and open it.
   If the file already exists, bump its sequence number until it doesn't.
   Open it with mode MODE ("w" or "wb", for text or binary files).
 */

static FILE *spoolcreate(name, type, host, gchar, seqno, mode)
char   *name;
int     type;
char   *host;
int     gchar;
char   *seqno;
char   *mode;
{
    FILE   *fp;
    char   *cp;

    /*
     * TODO: keep a record of files created by spoolcreate(), so that they can
     * be cleaned up by the error() function. While building the record, we
     * should use copies of the file names.
     */

    for (;;) {
	uucpname(name, type, host, gchar, seqno);
	cp = munge_filename(name);
	fp = fopen(cp, "r");
	if (fp == 0) break;		/* File doesn't exist */
	DEBUG(4, "Existing file named %s, skipping\n", cp);
	(void) fclose(fp);
	seqno = getseqnum();
    } 
    fp = fopen(cp, mode);
    if (fp == 0)
	error(EX_CANTCREAT, "%s: spool file open failed", cp);
    DEBUG(4, "create: %s\n", cp);
    return (fp);
}

/*
 * spoolmove - rename spoolcreate()-ed files 
 *
 * "from" is the old file name.  "to" is where we put the new one.
 * The rest of the args are used to build the new name.
 *
 * We allocate space for <old> but just a pointer for <new> because
 * the result of munge_filename is static, so we can only use the
 * most recent result "in place".
 */
static void spoolmove(from, to, type, host, gchar, seqno)
char   *from;
char   *to;
int     type;
char   *host;
int     gchar;
char   *seqno;
{
    char old[NAMESIZE];
    register char *new;
    FILE *fp;

    strcpy(old, munge_filename(from));
    /*
     * TODO: keep a record of files created by spoolcreate() and renamed by
     * spoolmove(), so that they can be cleaned up by the error() function.
     */

    for (;;) {
	uucpname(to, type, host, gchar, seqno);
	new = munge_filename(to);
	fp = fopen(new, "r");
	if (fp == 0) break;		/* File doesn't exist */
	DEBUG(4, "Existing file named %s, skipping\n", new);
	(void) fclose(fp);
	seqno = getseqnum();
    } 
    if (rename(old, new))
	error(EX_CANTCREAT, "%s: spool file rename from failed", old);
    DEBUG(4, "move: %s", old);
    DEBUG(4, " to: %s\n", new);
}

/* write_files - create spool files */

static void write_files()
{
    char    *seqno = getseqnum();	/* allocate sequence number */
    char    datafile[MAX_QFNAME];	/* D.hostSseqno */
    char    dxfile  [MAX_QFNAME];	/* D.hostXseqno */
    char    execfile[MAX_QFNAME];	/* X.hostXseqno */
    char    cmndfile[MAX_QFNAME];	/* C.hostGRADEseqno */
    char    tempfile[MAX_QFNAME];	/* T.hostTseqno */
    FILE   *fp;
    int     c;

    if (chdir(Spool)) {
	    fprintf(stderr, "Can't chdir to Spool dir: ");
	    perror(Spool);
	    exit(2);
    }

#ifdef	unix					/* what about other OSes? */
    (void) setgid(getegid());		
    (void) setuid(geteuid());
    (void) umask(022); /* Mask out group-write and other-write permission */
#endif

    /*
     * All spool files get the same sequence number. This is no problem as
     * long as we do not support host!file arguments or i/o redirection.
     */

    /*
     * Copy standard input to D (data) file if a "-" argument was specified.
     * I don't know why the data file gets a grade of 'S' here; on Sun uucp
     * (SunOS 4.0) it gets grade 'B'.  It doesn't seem to matter.
     */

    if (do_stdin) {
	fp = spoolcreate(datafile, 'D', host_name, 'S', seqno, "wb");
	while ((c = getchar()) != EOF)
	     if (putc(c, fp) == EOF) break;

	if (EOF == fclose(fp)) {
		/*
		 * Error in copying data to spool file.  E.g. disk full.
		 * We had better punt now, and return an error to caller.
		 */
		DEBUG(0, "Error in copying data to spool file %s\n", datafile);
		if (0 != remove(munge_filename(datafile))) {
			DEBUG(0, "Error in removing data spool file %s\n", datafile);
		}
		logit("XQT Q FAILED", xcmd);
		error(EX_CANTCREAT, "Can't copy data to spool file", "");
		/* NOTREACHED */
	}
    }

    /*
     * Create X (execute) file. This file should get a name X.hostXseqno on
     * the execution host. If the execution host is remote, we use the form
     * D.hostXseqno on the local host, and specify the final name in the S
     * request (see below). If the execution host is local, we use a
     * temporary name until the file is complete, and rename it afterwards.
     */

    if (remote_work) {
	fp = spoolcreate(dxfile, 'D', Myname, 'X', seqno, "w");
    } else {
	/*
	 * On multi-tasking OS's, some spool files should live under
	 * a temporary name while they are written. This may not be necessary on
	 * single-tasking systems, but then, it should not do any harm either.
	 */
	fp = spoolcreate(tempfile, 'T', host_name, 'T', seqno, "w");
    }
    (void) fprintf(fp, "U %s %s\n", who, Myname);
    if (nflag)
	(void) fprintf(fp, "N\n");
    if (zflag)
	(void) fprintf(fp, "Z\n");
    if (do_stdin) {
	(void) fprintf(fp, "F %s\n", datafile);
	(void) fprintf(fp, "I %s\n", datafile);
    }
    (void) fprintf(fp, "C %s\n", xcmd);
    (void) fclose(fp);		/* FIXME, check errors and punt */
    if (remote_work) {
	/* decide remote X.hostXnnnn name */
	uucpname(execfile, 'X', Myname, 'X', seqno);
    } else {
	/* make local X.mynameXnnnn */
	spoolmove(tempfile, execfile, 'X', Myname, 'X', seqno);
    }

    /*
     * Create a C (command) file for non-local command execution. For obvious
     * reasons, this file should be given a temporary name first, and it
     * should be renamed only when all other spool files have been finished.
     */

    if (remote_work) {
	fp = spoolcreate(tempfile, 'T', host_name, 'T', seqno, "w");
	if (do_stdin)
	    (void) fprintf(fp, "S %s %s %s - %s 0660\n",
			   datafile, datafile, who, datafile);
	(void) fprintf(fp, "S %s %s %s - %s 0660\n",
		       dxfile, execfile, who, dxfile);
	(void) fclose(fp);	/* FIXME, check errors and punt */
	spoolmove(tempfile, cmndfile, 'C', host_name, grade, seqno);
    }

    logit("XQT QUE'D", xcmd);
}

/* error - complain and die */

static void error(err, fmt, arg) 
	int err;
	char *fmt, *arg;
{
    /*
     * TODO: remove any files that were created by spoolcreate(), or renamed by
     * spoolmove(). 
     */

    (void) fputs("uux: ", stderr);
    (void) fprintf(stderr, fmt, arg);
    (void) putc('\n', stderr);
    exit(err);
}

/* usage - print usage message and die */

static void usage()
{
    error(EX_USAGE,
       "usage: uux [-gGRADE] [-n] [-xLEVEL] [-z] [-] [host!]commandstring", "");
}
